'use strict';

const fs = require('fs');
const url = require('url');
const { promisify } = require('util');
const request = promisify(require('request'));
const { getDefaultCacheDir } = require('./cache-utils');

async function getConfig({
  config,
  token
}) {
  if (typeof config !== 'object') {
    let isUrl = url.parse(config).host;

    if (isUrl) {
      config = (await request({
        url: config,
        headers: {
          Authorization: `Basic ${token}`
        }
      })).body;
    } else {
      config = fs.readFileSync(config);
    }

    config = JSON.parse(config);
  }

  if (config.cachePath === undefined) {
    config.cachePath = getDefaultCacheDir();
  }
  if (config.deleteBranch === undefined) {
    config.deleteBranch = true;
  }

  return config;
}

function _merge(left, right, normalize) {
  let config = Object.assign({}, left, right);
  let leftReviewers = typeof left.reviewers === 'string' ? left.reviewers : left.reviewers || [];
  let rightReviewers = typeof right.reviewers === 'string' ? right.reviewers : right.reviewers || [];
  if (Array.isArray(leftReviewers) && Array.isArray(rightReviewers) && normalize) {
    config.reviewers = [...new Set([
      ...leftReviewers,
      ...rightReviewers
    ])];
  }
  return config;
}

function _mergePackages(left, right) {
  let names = new Set([...Object.keys(left), ...Object.keys(right)]);

  let packages = {};
  for (let name of names) {
    packages[name] = _merge(
      left[name] || {},
      right[name] || {}
    );
  }
  return packages;
}

function _mergeConfigs(left, right, propToDelete) {
  let config = _merge(left, right, true);
  config.packages = _mergePackages(
    left.packages || {},
    right.packages || {}
  );
  delete config[propToDelete];
  return config;
}

function mergeConfigForRepo(config, repo) {
  return _mergeConfigs(config, repo, 'repositories');
}

function mergeConfigForPackage(config, name) {
  return _mergeConfigs(config, config.packages && config.packages[name] || {}, 'packages');
}

module.exports.getConfig = getConfig;
module.exports.mergeConfigForRepo = mergeConfigForRepo;
module.exports.mergeConfigForPackage = mergeConfigForPackage;
