'use strict';

const replaceTemplate = require('./replace-template');
const utils = require('./utils');
const repoUtils = require('./repo-utils');
const templateUtils = require('./template-utils');

async function createBranchAndPullRequest1({
  name,
  oldRange,
  newRange,
  token,
  project,
  repo,
  repoPath,
  prs,
  asyncRequests,
  endpoint,
  user,
  versionHint,
  branchTemplate = 'bb/{{package}}_{{newVersion}}',
  titleTemplate = '{{package}} {{oldRange}} => {{newRange}}',
  descriptionTemplate = 'update [{{package}}]({{url}}) {{oldRange}} => {{newRange}}\nversion {{oldVersion}} to {{newVersion}}',
  messageTemplate = '{{user}} updated {{package}} to {{newVersion}}',
  urlTemplate,
  reviewers,
  purgeStale,
  autoRebase,
  oldVersion,
  newVersion,
  registry,
  deleteBranch,
  shouldResetPackageJson,
  isMasterUpToDate,
  retainOriginalReviewers,
  preBranchHook,
  postBranchHook,
  prePrHook,
  postPrHook
}) {
  let templateValues = {
    endpoint,
    user,
    project,
    repo,
    package: name,
    oldRange,
    newRange,
    oldVersion,
    newVersion
  };

  let branch = replaceTemplate(branchTemplate, templateValues);

  templateValues = await module.exports.populateNewTemplateValues({
    branch,
    templateValues,
    repoPath,
    urlTemplate,
    name
  });

  let title = replaceTemplate(titleTemplate, templateValues);
  let description = replaceTemplate(descriptionTemplate, templateValues);

  let {
    shouldContinue,
    prToRebase
  } = await utils.processOldPrs({
    prs,
    branch,
    title,
    description,
    isMasterUpToDate,
    token,
    repoPath,
    autoRebase,
    purgeStale,
    name,
    newVersion,
    registry,
    branchTemplate,
    templateValues,
    versionHint,
    oldRange,
    deleteBranch
  });

  if (!shouldContinue) {
    return;
  }

  await repoUtils.updateDep({
    name,
    newRange,
    repoPath,
    registry,
    shouldResetPackageJson
  });

  await module.exports.createBranchAndPr({
    token,
    project,
    repo,
    repoPath,
    asyncRequests,
    endpoint,
    branch,
    messageTemplate,
    title,
    description,
    reviewers,
    prToRebase,
    templateValues,
    retainOriginalReviewers,
    preBranchHook,
    postBranchHook,
    prePrHook,
    postPrHook
  });
}

async function createBranchAndPullRequest2({
  branchTemplate = 'bb/update-deps',
  messageTemplate = '{{user}} updated dependencies',
  titleTemplate = 'update deps',
  descriptionTemplate = 'updated dependencies\n{{packages}}',
  urlTemplate,
  reviewers,
  purgeStale,
  autoRebase,
  meta,
  user,
  project,
  repo,
  prs,
  token,
  repoPath,
  deleteBranch,
  registry,
  asyncRequests,
  endpoint,
  retainOriginalReviewers,
  preBranchHook,
  postBranchHook,
  prePrHook,
  postPrHook
}) {
  let isMasterUpToDate = Object.values(meta).every(({ isMasterUpToDate }) => isMasterUpToDate);

  let templateValues = {
    endpoint,
    user,
    project,
    repo
  };

  let branch = replaceTemplate(branchTemplate, templateValues);

  templateValues = await module.exports.populateNewTemplateValues({
    branch,
    templateValues,
    singlePr: true,
    meta,
    repoPath,
    urlTemplate
  });

  let title = replaceTemplate(titleTemplate, templateValues);
  let description = replaceTemplate(descriptionTemplate, templateValues);

  let {
    shouldContinue,
    prToRebase
  } = await utils.processOldPrs({
    newVersions: meta,
    prs,
    branch,
    title,
    description,
    token,
    repoPath,
    autoRebase,
    purgeStale,
    deleteBranch,
    isMasterUpToDate,
    singlePr: true
  });

  if (!shouldContinue) {
    return;
  }

  for (let name of Object.keys(meta)) {
    let {
      newRange,
      isMasterUpToDate,
      shouldResetPackageJson
    } = meta[name];

    if (!isMasterUpToDate) {
      await repoUtils.updateDep({
        name,
        newRange,
        repoPath,
        registry,
        shouldResetPackageJson
      });
    }
  }

  await module.exports.createBranchAndPr({
    token,
    project,
    repo,
    repoPath,
    asyncRequests,
    endpoint,
    branch,
    messageTemplate,
    title,
    description,
    reviewers,
    prToRebase,
    templateValues,
    retainOriginalReviewers,
    preBranchHook,
    postBranchHook,
    prePrHook,
    postPrHook
  });
}

async function populateNewTemplateValues({
  branch,
  templateValues,
  singlePr,
  meta,
  repoPath,
  urlTemplate,
  name
}) {
  templateValues = Object.assign({}, templateValues, { branch });

  if (singlePr) {
    let packages = await templateUtils.getPackagesTemplateValue({
      meta,
      repoPath
    });
    // refrain from mutating because it's sent to other async functions
    templateValues = Object.assign({}, templateValues, { packages });
  }

  let packageUrl = await templateUtils.getUrlFromTemplate({
    shouldCheckRegistry: !singlePr,
    urlTemplate,
    templateValues,
    name,
    repoPath
  });
  // refrain from mutating because it's sent to other async functions
  templateValues = Object.assign({}, templateValues, { url: packageUrl });

  return templateValues;
}

async function createBranchAndPullRequest({
  token,
  project,
  repo,
  repoPath,
  asyncRequests = [],
  endpoint,
  branch,
  messageTemplate,
  title,
  description,
  reviewers,
  prToRebase,
  templateValues,
  retainOriginalReviewers,
  preBranchHook,
  postBranchHook,
  prePrHook,
  postPrHook
}) {
  if (preBranchHook) {
    await utils.callbackHook({
      url: preBranchHook,
      templateValues
    });
  }

  await repoUtils.createBranch({
    branch,
    repoPath,
    message: replaceTemplate(messageTemplate, templateValues)
  });

  let prPromise = Promise.resolve().then(async () => { // eslint-disable-line space-before-function-paren
    if (postBranchHook) {
      await utils.callbackHook({
        url: postBranchHook,
        templateValues
      });
    }

    if (prePrHook) {
      await utils.callbackHook({
        url: prePrHook,
        templateValues
      });
    }

    return await utils.createPr({
      token,
      endpoint,
      branch,
      project,
      repo,
      title,
      description,
      reviewers,
      prToRebase,
      retainOriginalReviewers,
      templateValues
    });
  }).then(async res => {
    if (postPrHook) {
      await utils.callbackHook({
        url: postPrHook,
        templateValues
      });
    }

    return res;
  });

  asyncRequests.push(prPromise);
}

module.exports.createBranchAndPr1 = createBranchAndPullRequest1;
module.exports.createBranchAndPr2 = createBranchAndPullRequest2;
module.exports.createBranchAndPr = createBranchAndPullRequest;
module.exports.populateNewTemplateValues = populateNewTemplateValues;
