'use strict';

const fs = require('fs');
const url = require('url');
const { promisify } = require('util');
const rimraf = promisify(require('rimraf'));
const path = require('path');
const run = require('./run');
const getPrs = require('./get-pull-requests');
const packageUtils = require('./package-utils');
const getRepoPath = require('./get-repo-path');
const {
  mergeConfigForRepo,
  mergeConfigForPackage
} = require('./config-utils');
const access = promisify(fs.access);
const mkdir = promisify(fs.mkdir);
const readFile = promisify(fs.readFile);
const {
  createBranchAndPr1,
  createBranchAndPr2
} = require('./create-branch-and-pull-request');

const {
  getProjectPackageVersion
} = packageUtils;

async function cloneRepo({
  token,
  endpoint,
  project,
  repo,
  cachePath
}) {
  let parsed = url.parse(endpoint);
  parsed.auth = Buffer.from(token, 'base64').toString();
  let _url = `${url.format(parsed)}scm/${project}/${repo}.git`;
  await run(`git clone ${_url}`, { cwd: cachePath });
}

async function gitResetClean(repoPath) {
  await run('git reset --hard', { cwd: repoPath });
  await run('git clean -fd', { cwd: repoPath });
}

async function exists(path) {
  try {
    await access(path, fs.constants.F_OK);
    return true;
  } catch (err) {
    return false;
  }
}

async function initCache(cachePath) {
  if (!await exists(cachePath)) {
    await mkdir(cachePath);
  }
}

async function initRepo({
  token,
  endpoint,
  cachePath,
  project,
  repo,
  repoPath
}) {
  if (!await exists(repoPath)) {
    return await cloneRepo({
      token,
      endpoint,
      project,
      repo,
      cachePath
    });
  }

  // solve a line-endings issues where the following pull would fail
  // because git status returns:
  // modified:   package.json
  // modified:   yarn.lock
  // await run('git checkout .', { cwd: repoPath });

  await run('git checkout master', { cwd: repoPath });

  await gitResetClean(repoPath);

  await run('git pull', { cwd: repoPath });
  await run('git fetch --prune', { cwd: repoPath });
}

function isValidPackage(name, packages) {
  if (Object.keys(packages).length === 0 || packages[name]) {
    return true;
  }
}

// yarn and npm aren't respecting local .npmrc's
// when they are also used to run the tests
// running mocha by itself doesn't suffer the problem
// this is a hack to use the local registry instead of bitbucket-npm-bot's registry
async function getRegistryHack(repoPath) {
  if (!process.env.npm_config_registry) {
    // not running inside npm or yarn,
    // so this hack is not necessary
    return;
  }

  let configPath = path.resolve(repoPath, '.npmrc');
  if (!await exists(configPath)) {
    return;
  }

  let config = await readFile(configPath, 'utf8');
  let [, registry] = config.match(/^registry = (.*)$/m) || [];
  return registry;
}

const depTypes = [
  'dependencies',
  'devDependencies'
];

const getVersionRange = require('./get-version-range');
const semver = require('semver');

async function convertPackageJsonToMeta({
  repoPath,
  packages,
  localConfig,
  registry
}) {
  let packageJson = require(path.join(repoPath, 'package.json'));
  let meta = {};

  for (let type of depTypes) {
    let deps = packageJson[type];
    for (let name in deps) {
      if (!isValidPackage(name, packages)) {
        continue;
      }

      let oldRange = deps[name];

      let packageConfig = mergeConfigForPackage(localConfig, name);

      let {
        versionHint,
        lockfileOnly = false
      } = packageConfig;

      let [oldVersion, newVersion] = await Promise.all([
        getProjectPackageVersion({
          name,
          repoPath
        }),
        packageUtils.getNewPackageVersion({
          name,
          repoPath,
          registry
        })
      ]);

      let newRange = getVersionRange(versionHint, oldRange, newVersion);

      let isMasterUpToDate = oldVersion === newVersion;

      let shouldResetPackageJson = lockfileOnly && semver.satisfies(newVersion, oldRange);

      meta[name] = {
        oldRange,
        oldVersion,
        newRange,
        newVersion,
        packageConfig,
        isMasterUpToDate,
        shouldResetPackageJson
      };
    }
  }

  return meta;
}

async function processDeps({
  asyncRequests,
  singlePr,
  localConfig,
  meta,
  registry,
  repoPath,
  deleteBranch,
  user,
  project,
  token,
  repo,
  prs,
  endpoint
}) {
  if (singlePr) {
    let {
      branchTemplate,
      messageTemplate,
      titleTemplate,
      descriptionTemplate,
      urlTemplate,
      reviewers,
      autoRebase,
      purgeStale,
      retainOriginalReviewers,
      preBranchHook,
      postBranchHook,
      prePrHook,
      postPrHook
    } = localConfig;

    await createBranchAndPr2({
      branchTemplate,
      titleTemplate,
      descriptionTemplate,
      messageTemplate,
      urlTemplate,
      reviewers,
      autoRebase,
      purgeStale,
      meta,
      user,
      project,
      repo,
      prs,
      token,
      repoPath,
      deleteBranch,
      registry,
      asyncRequests,
      endpoint,
      retainOriginalReviewers,
      preBranchHook,
      postBranchHook,
      prePrHook,
      postPrHook
    });
  } else {
    for (let name of Object.keys(meta)) {
      let {
        oldRange,
        oldVersion,
        newRange,
        newVersion,
        packageConfig: {
          endpoint,
          user,
          branchTemplate,
          purgeStale,
          autoRebase,
          titleTemplate,
          descriptionTemplate,
          messageTemplate,
          urlTemplate,
          versionHint,
          reviewers,
          retainOriginalReviewers,
          preBranchHook,
          postBranchHook,
          prePrHook,
          postPrHook
        },
        isMasterUpToDate,
        shouldResetPackageJson
      } = meta[name];

      await createBranchAndPr1({
        name,
        oldRange,
        newRange,
        token,
        project,
        repo,
        repoPath,
        prs,
        asyncRequests,
        endpoint,
        user,
        versionHint,
        branchTemplate,
        titleTemplate,
        descriptionTemplate,
        messageTemplate,
        urlTemplate,
        reviewers,
        purgeStale,
        autoRebase,
        oldVersion,
        newVersion,
        registry,
        deleteBranch,
        shouldResetPackageJson,
        isMasterUpToDate,
        retainOriginalReviewers,
        preBranchHook,
        postBranchHook,
        prePrHook,
        postPrHook
      });
    }
  }
}

async function processRepo({
  token,
  cachePath,
  repoPath,
  registry,
  repoConfig,
  project,
  repo,
  retries = 0
}) {
  let {
    endpoint,
    deleteBranch
  } = repoConfig;

  try {
    await initRepo({
      token,
      endpoint,
      cachePath,
      project,
      repo,
      repoPath
    });

    let localConfig = path.join(repoPath, '.bb-npm.js');
    localConfig = await exists(localConfig) ? require(localConfig) : {};

    localConfig = mergeConfigForRepo(repoConfig, localConfig);

    let {
      user,
      packages,
      singlePr
    } = localConfig;

    let prs = await getPrs({
      project,
      repo,
      token,
      endpoint,
      user
    });

    let meta = await convertPackageJsonToMeta({
      repoPath,
      packages,
      localConfig,
      registry
    });

    let asyncRequests = [];

    await processDeps({
      asyncRequests,
      singlePr,
      localConfig,
      meta,
      registry,
      repoPath,
      deleteBranch,
      user,
      project,
      token,
      repo,
      prs,
      endpoint
    });

    return {
      prs: await Promise.all(asyncRequests),
      retries
    };
  } catch (err) {
    if (retries > 0) {
      throw err;
    }

    await rimraf(repoPath);

    return await processRepo({
      token,
      cachePath,
      repoPath,
      registry,
      repoConfig,
      project,
      repo,
      retries: ++retries
    });
  }
}

async function _run({
  config,
  token
}) {
  let {
    cachePath,
    repositories,
    disabled
  } = config;

  if (disabled) {
    return;
  }

  await initCache(cachePath);

  repositories = repositories.filter(repo => !repo.skip);
  let only = repositories.filter(repo => repo.only);
  if (only.length) {
    repositories = only;
  }

  return {
    repos: (await Promise.all(repositories.map(async repo => {
      let repoPath = getRepoPath(cachePath, repo.name);
      let registry = await getRegistryHack(repoPath);

      return {
        repo,
        result: await processRepo({
          token,
          cachePath,
          repoPath,
          registry,
          repoConfig: mergeConfigForRepo(config, repo),
          project: repo.project,
          repo: repo.name
        })
      };
    }))).reduce((output, { repo, result }) => {
      output[`${repo.project}/${repo.name}`] = result;
      return output;
    }, {}),
    cachePath
  };
}

module.exports = _run;
module.exports.initCache = initCache;
module.exports.initRepo = initRepo;
module.exports.getRegistryHack = getRegistryHack;
module.exports.getRepoPath = getRepoPath;
