'use strict';

const run = require('./run');
const runWithRegistry = require('./run-with-registry');
const path = require('path');
const { promisify } = require('util');
const rimraf = promisify(require('rimraf'));
const fs = require('fs');
const access = promisify(fs.access);

async function getPackageVersions({
  name,
  repoPath,
  registry
}) {
  // let versions = JSON.parse(await runWithRegistry(`yarn info ${name} versions --json`, registry, { cwd: repoPath })).data;
  let versions = JSON.parse(await runWithRegistry(`npm info ${name} versions --json`, registry, { cwd: repoPath }));
  return versions;
}

async function getNewPackageVersion({
  name,
  repoPath,
  registry
}) {
  // let newVersion = JSON.parse(await runWithRegistry(`yarn info ${name} version --json`, registry, { cwd: repoPath })).data;
  let newVersion = JSON.parse(await runWithRegistry(`npm info ${name} version --json`, registry, { cwd: repoPath }));
  return newVersion;
}

async function getProjectPackageVersion({
  branch = 'master',
  name,
  repoPath
}) {
  return (await getProjectPackageVersions({
    branch,
    names: [name],
    repoPath
  }))[name];
}

async function exists(file) {
  try {
    await access(file, fs.constants.F_OK);
    return true;
  } catch (err) {
    return false;
  }
}

async function getProjectPackageVersions({
  branch = 'master',
  names,
  repoPath
}) {
  try {
    await run(`git checkout -B ${branch} origin/${branch}`, { cwd: repoPath });
  } catch (err) {
    // branch was deleted from PR
    return {};
  }

  let isYarn = await exists(path.resolve(repoPath, 'yarn.lock'));

  let json;
  async function getVersion(name) {
    if (!json) {
      if (isYarn) {
        // `yarn list --depth 0` isn't optimal yet
        // https://github.com/yarnpkg/yarn/issues/3569
        let depsString = await run('yarn list --depth 0 --json', { cwd: repoPath });
        json = JSON.parse(depsString);
      } else {
        await run('npm install', { cwd: repoPath });

        let depsString;
        try {
          depsString = await run('npm ls --depth 0 --json', { cwd: repoPath });
        } catch (err) {
          if (err.stderr.includes('npm ERR! extraneous')) {
            depsString = err.stdout;
          } else {
            throw err;
          }
        }
        json = JSON.parse(depsString);
      }
    }

    let version;
    if (isYarn) {
      version = json.data.trees.map(({ name }) => {
        // handle scoped packages ie. "@ember/optional-features@^0.7.0"
        let i = name.lastIndexOf('@');
        return {
          name: name.substr(0, i),
          version: name.substr(i + 1)
        };
      }).find(dep => dep.name === name).version;
    } else {
      version = json.dependencies[name].version;
    }
    return version;
  }

  let versions = {};
  for (let name of names) {
    let version;
    try {
      version = await getVersion(name);
    } catch (err) {
      // yarn bug doesn't update modules sometimes
      await rimraf(path.resolve(repoPath, 'node_modules', name));
      await rimraf(path.resolve(repoPath, 'node_modules/.yarn-integrity'));

      version = await getVersion(name);
    }
    versions[name] = version;
  }

  await run('git checkout master', { cwd: repoPath });
  if (branch !== 'master') {
    await run(`git branch -D ${branch}`, { cwd: repoPath });
  }

  return versions;
}

module.exports.getPackageVersions = getPackageVersions;
module.exports.getNewPackageVersion = getNewPackageVersion;
module.exports.getProjectPackageVersion = getProjectPackageVersion;
module.exports.getProjectPackageVersions = getProjectPackageVersions;
