'use strict';

const packageUtils = require('./package-utils');
const repoUtils = require('./repo-utils');
const prUtils = require('./pr-utils');
const utils = require('./utils');
const replaceTemplate = require('./replace-template');
const getVersionRange = require('./get-version-range');

async function processOldPullRequests({
  newVersions,
  prs,
  branch,
  title,
  description,
  isMasterUpToDate,
  token,
  repoPath,
  autoRebase,
  purgeStale,
  name,
  newVersion,
  registry,
  branchTemplate,
  templateValues,
  versionHint,
  oldRange,
  deleteBranch,
  singlePr
}) {
  let matchedOlderVersionBranches;
  let prToRebase = null;
  let hasUpToDatePr = false;
  let declines = [];

  function decline(pr) {
    if (pr.state !== 'DECLINED') {
      declines.push(utils.declinePr({
        token,
        repoPath,
        pr,
        deleteBranch
      }));
    }
  }

  function rebase(pr) {
    if (pr.state !== 'DECLINED') {
      prToRebase = pr;
    }
  }

  for (let pr of prs) {
    if (pr.fromRef.displayId === branch) {
      if (pr.state === 'DECLINED' && pr.title === title && pr.description === description) {
        hasUpToDatePr = true;
        continue;
      }

      if (isMasterUpToDate) {
        decline(pr);
        continue;
      }

      if (autoRebase) {
        let _isRebaseNeeded = await repoUtils.isRebaseNeeded({
          branch,
          repoPath
        });
        if (_isRebaseNeeded) {
          rebase(pr);
          continue;
        }
      }

      if (autoRebase || purgeStale) {
        if (!singlePr) {
          newVersions = { [name]: { newVersion } };
        }
        let branchVersions = await packageUtils.getProjectPackageVersions({
          branch,
          names: Object.keys(newVersions),
          repoPath
        });
        let sameBranchOldVersion = Object.keys(branchVersions).some(name => branchVersions[name] !== newVersions[name].newVersion);
        if (sameBranchOldVersion) {
          if (autoRebase) {
            rebase(pr);
            continue;
          }
          if (purgeStale) {
            decline(pr);
            continue;
          }
        }
      }

      let conflicted = await prUtils.isPrConflicted({
        pr,
        token
      });
      if (conflicted) {
        if (purgeStale) {
          decline(pr);
          continue;
        }
        rebase(pr);
        continue;
      }

      if (pr.state !== 'DECLINED') {
        hasUpToDatePr = true;
      }
    } else {
      if (!matchedOlderVersionBranches) {
        matchedOlderVersionBranches = new Set();
        if (purgeStale && !singlePr) {
          let versions = await packageUtils.getPackageVersions({
            name,
            repoPath,
            registry
          });

          for (let version of versions) {
            if (version === newVersion) {
              continue;
            }
            let branch = replaceTemplate(
              branchTemplate,
              Object.assign({}, templateValues, {
                newVersion: version,
                newRange: getVersionRange(versionHint, oldRange, version)
              })
            );
            matchedOlderVersionBranches.add(branch);
          }
        }
      }

      if (matchedOlderVersionBranches.has(pr.fromRef.displayId)) {
        decline(pr);
      }
    }
  }

  await Promise.all(declines);

  return {
    shouldContinue: !(isMasterUpToDate || hasUpToDatePr),
    prToRebase
  };
}

module.exports = processOldPullRequests;
